﻿using System;
using Benday.FakesSample.Business;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Benday.FakesSample.DataAccess;
using Microsoft.QualityTools.Testing.Fakes;
using Benday.FakesSample.Common;

namespace Benday.FakesSample.UnitTests
{
    [TestClass]
    public class PersonManagerFixture
    {
        private readonly DateTime EXPECTED_DEFAULT_DATE = 
            new DateTime(1900, 1, 1);

        [TestInitialize]
        public void OnTestInitialize()
        {
            
        }

        [TestCleanup]
        public void OnTestCleanup()
        {
            _SystemUnderTest = null;
            _LoggerInstance = null;
            _RepositoryInstance = null;
        }

        private PersonManager _SystemUnderTest;
        public PersonManager SystemUnderTest
        {
            get
            {
                if (_SystemUnderTest == null)
                {
                    _SystemUnderTest = 
                        new PersonManager(
                            RepositoryInstance, 
                            LoggerInstance);
                }

                return _SystemUnderTest;
            }
        }

        private IRepository<IPerson> _RepositoryInstance;
        public IRepository<IPerson> RepositoryInstance
        {
            get
            {
                if (_RepositoryInstance == null)
                {
                    _RepositoryInstance = 
                        new RepositoryMock<IPerson>();
                }
                return _RepositoryInstance;
            }
        }

        private Benday.FakesSample.Common.Fakes.StubILogger _LoggerInstance;
        public Benday.FakesSample.Common.Fakes.StubILogger LoggerInstance
        {
            get
            {
                if (_LoggerInstance == null)
                {
                    _LoggerInstance = new Benday.FakesSample.Common.Fakes.StubILogger();
                }
                return _LoggerInstance;
            }
        }

        [TestMethod]
        public void WhenNewRecordIsCreatedButNotSavedCreateDateIs_19000101()
        {
            var person = SystemUnderTest.CreateNew();

            Assert.AreEqual<DateTime>(
                new DateTime(1900, 1, 1), 
                person.CreateDate, 
                "CreateDate was wrong.");
        }

        [TestMethod]
        public void WhenNewRecordIsCreatedButNotSavedLastModifiedIs_19000101()
        {
            var person = SystemUnderTest.CreateNew();

            Assert.AreEqual<DateTime>(
                EXPECTED_DEFAULT_DATE,
                person.LastModified,
                "LastModified was wrong.");
        }

        [TestMethod]
        public void WhenNewRecordIsCreatedButNotSavedIdIsLessThan1()
        {
            var person = SystemUnderTest.CreateNew();

            Assert.IsTrue(
                person.Id < 1, 
                "Person id should be less than 1.  Actual value was '{0}'.", 
                person.Id);            
        }

        private IPerson CreateAndPopulatePerson(
            string firstName, 
            string lastName, 
            string emailAddress, 
            string notes)
        {
            var person = SystemUnderTest.CreateNew();

            person.FirstName = firstName;
            person.LastName = lastName;
            person.Notes = notes;
            person.EmailAddress = emailAddress;

            return person;
        }

        [TestMethod]
        public void WhenNewRecordIsSavedCreateDateAndLastModifiedIsUpdated()
        {
            using (ShimsContext.Create())
            {
                DataAccess.Fakes.ShimPersonDataAccess.AllInstances.SavePersonFakesDataSetPersonRow
                    = (x, y) => { };

                var person =
                    CreateAndPopulatePerson(
                        "Ben", "Day",
                        "benday@benday.com", "yay! notes!!!");

                SystemUnderTest.Save(person);

                /*
                Assert.IsTrue(person.Id > 0, 
                    "Person Id should be greater than 0.  Actual value was '{0}'.",
                    person.Id);
                */

                Assert.AreNotEqual<DateTime>(
                    EXPECTED_DEFAULT_DATE,
                    person.LastModified,
                    "LastModified was wrong.");

                Assert.AreNotEqual<DateTime>(
                    EXPECTED_DEFAULT_DATE,
                    person.CreateDate,
                    "CreateDate was wrong.");
            }
        }

        private IPerson GetSavedIPerson()
        {
            IPerson person = new Benday.FakesSample.Common.Fakes.StubIPerson();

            person.Id = 123;
            person.FirstName = "Ben";
            person.LastName = "Day";
            person.EmailAddress = "benday@benday.com";
            person.Notes = "notes!";
            person.CreateDate = new DateTime(2012, 3, 12);
            person.LastModified = new DateTime(2012, 4, 14);

            return person;
        }

        [TestMethod]
        public void WhenExistingRecordIsSavedCreateDateAndLastModifiedIsUpdated()
        {
            IPerson person = GetSavedIPerson();

            DateTime originalCreateDate = person.CreateDate;
            DateTime originalLastModified = person.LastModified;

            person.Notes = "wheeeeeee!!! modified notes!!!";
            SystemUnderTest.Save(person);

            Assert.AreEqual<DateTime>(
                originalCreateDate,
                person.CreateDate,
                "Create date should not change.");

            Assert.AreNotEqual<DateTime>(
                originalLastModified,
                person.LastModified,
                "Last modified was not updated.");

            /*
            using (ShimsContext.Create())
            {
                DataAccess.Fakes.ShimPersonDataAccess.AllInstances.SavePersonFakesDataSetPersonRow
                    = (x, y) => { };

                DataAccess.Fakes.ShimPersonDataAccess.AllInstances.LoadPersonInt32 =
                    (x, personId) =>
                    {
                        FakesDataSet.PersonRow row = 
                            new FakesDataSet().Person.NewPersonRow();

                        row.Id = 123;
                        row.FirstName = "Ben";
                        row.LastName = "Day";
                        row.EmailAddress = "benday@benday.com";
                        row.Notes = "notes!";
                        row.CreateDate = new DateTime(2012, 3, 12);
                        row.LastModified = new DateTime(2012, 4, 14);

                        return row;
                    };

                var person = SystemUnderTest.LoadById(123);

                DateTime originalCreateDate = person.CreateDate;
                DateTime originalLastModified = person.LastModified;

                person.Notes = "wheeeeeee!!! modified notes!!!";
                SystemUnderTest.Save(person);

                Assert.AreEqual<DateTime>(
                    originalCreateDate,
                    person.CreateDate,
                    "Create date should not change.");

                Assert.AreNotEqual<DateTime>(
                    originalLastModified,
                    person.LastModified,
                    "Last modified was not updated.");
            }
            */
        }

        [TestMethod]
        [ExpectedException(typeof(SqlServerOutOfDiskSpaceException))]
        public void VerifySqlServerOutOfDiskSpaceIsHandled()
        {
            var person = new Benday.FakesSample.Common.Fakes.StubIPerson();

            person.IdGet = () =>
            {
                throw new SqlServerOutOfDiskSpaceException("No more room!");
            };

            SystemUnderTest.Save(person);
        }

        [TestMethod]
        [ExpectedException(typeof(SqlServerOutOfDiskSpaceException))]
        public void VerifyLoggerIsCalledWhenSqlServerOutOfDiskSpaceIsHandled()
        {
            var person = new Benday.FakesSample.Common.Fakes.StubIPerson();

            person.IdGet = () =>
            {
                throw new SqlServerOutOfDiskSpaceException("No more room!");
            };

            var wasWriteToLogCalled = false;

            LoggerInstance.WriteToLogException = (ex) =>
            {
                wasWriteToLogCalled = true;
            };

            try
            {
                SystemUnderTest.Save(person);    
            }
            catch (Exception ex)
            {
                Assert.IsTrue(wasWriteToLogCalled, "Logger was not called.");
                throw ex;
            }            
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void GetThatLastBitOfCodeCoverageFromSaveUsingAStub()
        {
            var person = new Benday.FakesSample.Common.Fakes.StubIPerson();

            person.IdGet = () =>
            {
                throw new ArgumentNullException("BOOOOOOOOM!");
            };

            SystemUnderTest.Save(person);
        }
        
        [TestMethod]
        public void SaveStubInstanceOfPerson()
        {
            IPerson person = new Benday.FakesSample.Common.Fakes.StubIPerson();

            person.FirstName = "Skip";
            person.LastName = "Rosenwinkle";
            person.EmailAddress = "skip@rosenwinkle.net";
            person.Notes = "yeehaw!";

            DateTime originalCreateDate = person.CreateDate;
            DateTime originalLastModified = person.LastModified;

            SystemUnderTest.Save(person);

            Assert.AreNotEqual<DateTime>(
                originalCreateDate,
                person.CreateDate,
                "Create date should change.");

            Assert.AreNotEqual<DateTime>(
                originalLastModified,
                person.LastModified,
                "Last modified should change.");

            /*
            using (ShimsContext.Create())
            {
                DataAccess.Fakes.ShimPersonDataAccess.AllInstances.SavePersonFakesDataSetPersonRow
                    = (x, y) => { };

                IPerson person = new Benday.FakesSample.Common.Fakes.StubIPerson();

                person.FirstName = "Skip";
                person.LastName = "Rosenwinkle";
                person.EmailAddress = "skip@rosenwinkle.net";
                person.Notes = "yeehaw!";

                DateTime originalCreateDate = person.CreateDate;
                DateTime originalLastModified = person.LastModified;

                SystemUnderTest.Save(person);

                Assert.AreNotEqual<DateTime>(
                    originalCreateDate,
                    person.CreateDate,
                    "Create date should change.");

                Assert.AreNotEqual<DateTime>(
                    originalLastModified,
                    person.LastModified,
                    "Last modified should change.");
            }
            */
        }
    }
}
